"""configuring different parts of the houses pipeline module"""
import pathlib
import pandas as pd
import houses_pipeline

pd.options.display.max_rows = 10
pd.options.display.max_columns = 10

PACKAGE_ROOT = pathlib.Path(houses_pipeline.__file__).resolve().parent
TRAINED_MODEL_DIR = PACKAGE_ROOT / "trained_models"
DATASET_DIR = PACKAGE_ROOT / "datasets"

# variables
FEATURES = [
    "MSSubClass",
    "MSZoning",
    "Neighborhood",
    "OverallQual",
    "OverallCond",
    "YearRemodAdd",
    "RoofStyle",
    "MasVnrType",
    "BsmtQual",
    "BsmtExposure",
    "HeatingQC",
    "CentralAir",
    "1stFlrSF",
    "GrLivArea",
    "BsmtFullBath",
    "KitchenQual",
    "Fireplaces",
    "FireplaceQu",
    "GarageType",
    "GarageFinish",
    "GarageCars",
    "PavedDrive",
    "LotFrontage",
    # this one is only to calculate temporal variable:
    "YrSold",
]

# this variable is to calculate the temporal variable,
# can be dropped afterwards
DROP_FEATURES = "YrSold"

# numerical variables with NA in train set
NUMERICAL_VARS_WITH_NA = ["LotFrontage"]

# categorical variables with NA in train set
CATEGORICAL_VARS_WITH_NA = [
    "MasVnrType",
    "BsmtQual",
    "BsmtExposure",
    "FireplaceQu",
    "GarageType",
    "GarageFinish",
]

TEMPORAL_VARS = "YearRemodAdd"

# variables to log transform
NUMERICALS_LOG_VARS = ["LotFrontage", "1stFlrSF", "GrLivArea"]

# categorical variables to encode
CATEGORICAL_VARS = [
    "MSZoning",
    "Neighborhood",
    "RoofStyle",
    "MasVnrType",
    "BsmtQual",
    "BsmtExposure",
    "HeatingQC",
    "CentralAir",
    "KitchenQual",
    "FireplaceQu",
    "GarageType",
    "GarageFinish",
    "PavedDrive",
]

NUMERICAL_NA_NOT_ALLOWED = [
    feature
    for feature in FEATURES
    if feature not in CATEGORICAL_VARS + NUMERICAL_VARS_WITH_NA
]

CATEGORICAL_NA_NOT_ALLOWED = [
    feature for feature in CATEGORICAL_VARS if feature not in CATEGORICAL_VARS_WITH_NA
]


PIPELINE_NAME = "lasso_regression"
PIPELINE_SAVE_FILE = f"{PIPELINE_NAME}_output_v"

# used for differential testing
ACCEPTABLE_MODEL_DIFFERENCE = 0.05



"""Train a lasso regression"""
# generally used modules
# import joblib
import pandas as pd
import click

# pipeline feature engineering and scaling
from sklearn.compose import TransformedTargetRegressor, ColumnTransformer
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import StandardScaler
from sklearn.preprocessing import PowerTransformer, OneHotEncoder
from sklearn.preprocessing import OrdinalEncoder
from sklearn.impute import SimpleImputer

# modeling
from sklearn.linear_model import Lasso

# model tracking
import mlflow
import mlflow.sklearn

# internal modules
from houses_pipeline import constants

from houses_pipeline.transformers import RareCategoriesReplacer

__ordinal_encoder = OrdinalEncoder(
    categories=[constants.ORDINALS_ORDERING] * len(constants.ORDINALS),
    handle_unknown='use_encoded_value',
    unknown_value=-1
)


@click.command()
@click.argument(
    'input_filepath',
    type=click.Path(exists=True),
    default=constants.DEFAULT_INPUT_PATH
)
@click.option(
    '--alpha',
    'alpha',
    required=False,
    default=0.05,
    help="Regularization parameter of the lasso regression"
)
@click.option(
    '--model_seed',
    'model_seed',
    required=False,
    default=0.05,
    help="Regularization model seed"
)
@click.option(
    '--rare_threshold',
    'rare_threshold',
    required=False,
    default=0.05,
    help="Rare Threshold"
)
def main(input_filepath, alpha, model_seed, rare_threshold):
    """Main method for training the lasso model"""
    train_df = pd.read_csv(input_filepath)


    # to set some additional notes in mlflow tracking use:
    # mlflow.note.content

    X_train = train_df.drop(constants.TARGET_VARIABLE_NAME, axis=1)
    y_train = train_df[constants.TARGET_VARIABLE_NAME]

    categoricals = X_train.columns & constants.CATEGORICAL_COLUMNS
    numericals = X_train.columns & constants.NUMERICAL_COLUMNS

    with mlflow.start_run():
        # transformations to apply to the numerical predictors
        numeric_pipeline = Pipeline([
            ('impute_with_mean', SimpleImputer(strategy='mean')),
            ('scaling', StandardScaler()),
            ('yeo_johnson', PowerTransformer())
        ])

        # transformations to be applied to categorical features
        categoric_pipeline = Pipeline([
            ('replace_rare', RareCategoriesReplacer(rare_threshold)),
            ('one_hot_encoding', OneHotEncoder())
        ])

        # transformations on the response variable
        response_variable_pipeline = Pipeline([
            ('normalization', StandardScaler()),
            ('yeo_johnson', PowerTransformer())
        ])

        all_transformations = ColumnTransformer(
            transformers=[
                ('numeriric_transformations', numeric_pipeline, numericals),
                ('categoric_transformations', categoric_pipeline, categoricals),
                ('ordinals_encoding', __ordinal_encoder, constants.ORDINALS)
            ],
            remainder='passthrough'
        )

        # track the experiment's transformations
        mlflow.log_param("rare_threshold", rare_threshold)
        mlflow.log_param("numeric_transformations", "norm, yj, impute_mean")
        mlflow.log_param("categorical_transformations", "replace_rare, ohe")
        mlflow.log_param("ordinals_transformations", "ordinal_encoding")

        pipeline = Pipeline(
            steps=[
                ('column_transformations', all_transformations),
                ('lasso_and_target_transform', TransformedTargetRegressor(
                    regressor=Lasso(alpha=alpha, random_state=model_seed),
                    transformer=response_variable_pipeline
                ))
            ]
        )

        # fit the pipeline
        pipeline.fit(X_train, y_train)

        # track model parameters and metrics
        mlflow.log_param('alpha', alpha)
        mlflow.log_param('model_random_seed', model_seed)

        mlflow.sklearn.log_model(
            sk_model=pipeline,
            artifact_path="lasso_model",
            code_paths=[
                "notebooks/03-ns-modelling.ipynb",
                "pipeline/modelling/train_lasso.py"
            ],
            registered_model_name="LassoRegression"
        )


if __name__ == "__main__":
    # pylint: disable=no-value-for-parameter
    main()



"""
The pipeline module containing extract, preprocess and train
"""
import logging

from houses_pipeline.config import config
from houses_pipeline.config import logging_configuration
from houses_pipeline.constants import PACKAGE_ROOT

VERSION_PATH = PACKAGE_ROOT / 'VERSION'

# Configure logger for use in package
logger = logging.getLogger(__name__)
# logger.setLevel(logging.DEBUG)
# logger.addHandler(logging_config.get_console_handler())
# logger.propagate = False


with open(VERSION_PATH, 'r', encoding='utf-8') as version_file:
    __version__ = version_file.read().strip()



"""
Preprocess the input data to generate a ready to use dataset
in the modelling step of the pipeline
"""

# general utilities
import logging
import pandas as pd
import numpy as np
import click

# sklearn pipeline modules
from sklearn.pipeline import Pipeline
from sklearn.impute import SimpleImputer
from sklearn.preprocessing import FunctionTransformer

# internal modules
from houses_pipeline.transformers import Pandalizer
from houses_pipeline import constants


@click.command()
@click.argument(
    'input_filepath',
    type=click.Path(exists=True),
    default=constants.DEFAULT_INPUT_PATH
)
@click.argument(
    'output_filepath',
    type=click.Path(),
    default=constants.DEFAULT_OUTPUT_PATH
)
@click.option(
    '-v',
    '--verbose',
    'verbose',
    required=False,
    default=False,
    help=constants.VERBOSE_HELP
)
def main(input_filepath: str, output_filepath: str, verbose: bool):
    """Preprocess the dataset and turn it from the given input to the output"""
    logger = logging.getLogger(__name__)
    # click.echo() or logger.info?
    logger.info("Transforming %s to %s", input_filepath, output_filepath)

    # load the input dataframe
    input_df = pd.read_csv(input_filepath)

    # create the preprocessing pipeline
    pipeline = Pipeline(
        steps=[
            ('drop_useless_columns', FunctionTransformer(
                lambda df: df.drop(constants.DROP_COLUMNS, axis=1))
            ),
            ('impute_missing_categories', Pandalizer(
                SimpleImputer(strategy='constant', fill_value='Missing')
            ))
            # Usually this pipeline might also have:
            # removing duplicate rows
            # formatting dates
            # transforming types and scales to be the same, unification, etc
            # removing erroneous data

            # I would consider it doesn't have anything learnable from
            # the dataset, or experimentable.
        ],
        verbose=verbose
    )

    # transform the data
    to_impute = np.append(constants.CATEGORICAL_COLUMNS, constants.ORDINALS)
    output_houses_dataframe = pipeline.fit_transform(
        input_df,
        impute_missing_categories__columns=to_impute
    )

    # save the output to the specified path
    output_houses_dataframe.to_csv(output_filepath, index=False)

if __name__ == '__main__':
    LOG_FORMAT = '%(asctime)s - %(name)s - %(levelname)s - %(message)s'
    logging.basicConfig(level=logging.INFO, format=LOG_FORMAT)

    # pylint: disable=no-value-for-parameter
    main()



"""
Test transformers
"""
from sklearn.impute import SimpleImputer
import pandas as pd
import numpy as np
from ..transformers import Pandalizer


def test_pandalizer_returns_dataframe():
    """Test whether the Pandalizer returns proper panda dataframe"""
    dataframe = pd.DataFrame({
        'x': [1, 2, 3, np.nan],
        'y': [np.nan, 12, 1.4,  5.6]
    })

    imputer = Pandalizer(SimpleImputer(strategy='mean'))
    result_dataframe = imputer.fit_transform(dataframe)
    assert isinstance(result_dataframe, pd.DataFrame)



"""Data Transformers for helping with the preprocessing step"""
from sklearn.base import BaseEstimator, TransformerMixin
import pandas as pd
import numpy as np

class Pandalizer(BaseEstimator, TransformerMixin):
    """
    Executes a transformer on a subset of columns and returns a
    Pandas DataFrame as a result
    """
    def __init__(self, transformer):
        self.transformer = transformer
        self.columns = None


    def fit(self, df, y=None, columns=None, **fit_args):
        """Fit the Pandalizer taking additional fit_args to pass"""
        # get only a subset of columns on which to apply the transformer
        if columns is None:
            self.columns = df.columns
        else:
            self.columns = np.intersect1d(columns, df.columns)
        return self.transformer.fit(df[self.columns], y, **fit_args)


    def transform(self, data):
        """Transform the data"""
        df = data.copy()
        df.loc[:, self.columns] = self.transformer.transform(df[self.columns])
        return df


    def fit_transform(self, X, y=None, columns=None, **fit_args):
        """Fit_Transform the data"""
        # pylint: disable=unused-argument
        self.fit(X, y, columns, **fit_args)
        return self.transform(X)


class RareCategoriesReplacer(BaseEstimator, TransformerMixin):
    """
    Replaces Categorical Columns rare values with a keyword
    """
    def __init__(self, threshold=0.05, keyword: str='Other') -> None:
        self.keyword = keyword
        self.threshold = threshold
        self.proportions = []


    def get_proportions(self, X):
        """
        Get the proportions of the keywords in the categorical columns
        """
        counts = [pd.Series(x).value_counts(normalize=True) for x in X.T]
        return counts


    def fit(self, X, y=None):
        """Fit the rare categorical transformer"""
        # pylint: disable=unused-argument
        is_df = isinstance(X, pd.DataFrame)
        self.proportions = self.get_proportions(X.values if is_df else X)
        return self


    def is_to_replace(self, i, col):
        """calculate keywords to replace by a given column"""
        props = self.proportions[i]
        rares = props[props < self.threshold].index.values
        new_ones = np.setdiff1d(col, props.index)
        is_rare = np.isin(col, rares)
        is_new_one = np.isin(col, new_ones)
        return is_rare | is_new_one


    def transform(self, X) -> pd.DataFrame:
        """Transform the rare categorical transformer"""
        is_df = isinstance(X, pd.DataFrame)
        Xt = X.values.copy() if is_df else X.values.copy()
        for i, col in enumerate(Xt.T):
            col[self.is_to_replace(i, col)] = self.keyword
        return Xt


class RedundantColumnsRemover(BaseEstimator, TransformerMixin):
    """
    Removes columns with a high proportion of repeating values
    """
    def __init__(self):
        self.columns_to_remove = None
        self.additional = None


    def _get_max_count(self, column):
        """Return the maximal value count for a categorical column"""
        category_props = column.value_counts(normalize=True)
        return category_props.max()


    def fit(self, X, y=None, threshold: float=0.9, additional: list[str]=None):
        """Learn the columns which are redundantly occupied"""
        # pylint: disable=unused-argument
        additional = [] if not additional else additional
        max_counts = X.apply(self._get_max_count, axis=0)
        above_threshold = max_counts[max_counts > threshold]
        self.columns_to_remove = above_threshold.index.values
        self.additional = additional
        return self


    def transform(self, X):
        """Drop the redundant columns plus additionally given ones"""
        columns_to_remove = np.append(
            self.columns_to_remove, self.additional
        )
        result = X.drop(columns_to_remove, axis=1)
        return result



"""setup our houses_pipeline package"""
#!/usr/bin/env python
# -*- coding: utf-8 -*-
import io
import os
from pathlib import Path
from setuptools import find_packages, setup

# Package meta-data.
NAME = 'houses_pipeline'
DESCRIPTION = "Houses pipeline for preprocessing and training a model"
URL = 'https://github.com/nshahpazov/houses-ds-project'
EMAIL = 'nshahpazov@gmail.com'
AUTHOR = 'Nikola Shahpazov'
REQUIRES_PYTHON = '>=3.9.2'


# def list_reqs(fname='requirements.txt'):
#     """Packages that are required for this module to be executed"""
#     with open(fname, encoding="utf-8") as descriptor:
#         return descriptor.read().splitlines()


here = os.path.abspath(os.path.dirname(__file__))

# Import the README and use it as the long-description.
# Note: this will only work if 'README.md' is present in your MANIFEST.in file!
try:
    with io.open(os.path.join(here, 'README.md'), encoding='utf-8') as file:
        LONG_DESCRIPTION = '\n' + file.read()
except FileNotFoundError:
    LONG_DESCRIPTION = DESCRIPTION


# Load the package's __version__.py module as a dictionary.
ROOT_DIR = Path(__file__).resolve().parent
PACKAGE_DIR = ROOT_DIR / 'houses_pipeline'
about = {}

with open(PACKAGE_DIR / 'VERSION', encoding='utf-8') as file:
    _version = file.read().strip()
    about['__version__'] = _version


# Where the magic happens:
setup(
    name=NAME,
    version=about['__version__'],
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type='text/markdown',
    author=AUTHOR,
    author_email=EMAIL,
    python_requires=REQUIRES_PYTHON,
    url=URL,
    packages=find_packages(exclude=('tests',)),
    # package_dir={'': PACKAGE_DIR},
    package_data={'houses_pipeline': ['VERSION']},
    # install_requires=list_reqs(),
    extras_require={
        "dev": ["pytest>=7.1.2"]
    },
    include_package_data=True,
    license='BSD 3',
    classifiers=[
        # Trove classifiers
        # Full list: https://pypi.python.org/pypi?%3Aaction=list_classifiers
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy'
    ],
)



"""Different constants used for the pipeline"""
import pathlib
import houses_pipeline

# the root of the package
PACKAGE_ROOT = pathlib.Path(houses_pipeline.__file__).resolve().parent

YEAR_COLUMNS = ['YearBuilt', 'YearRemodAdd', 'GarageYrBlt']
INTERVAL_COLUMNS = ['BuiltTilSold', 'RemodTilSold', 'GarageBuildTilSold']
# LISTWISE_DELETION_COLUMNS = ['MasVnrArea', 'MasVnrType', 'Electrical']
DEFAULT_MISSING_THRESHOLD = 0.80

REDUNDANT_COLUMNS = [
    'Street', 'Utilities', 'LandSlope', 'Condition2', 'RoofMatl', 'BsmtCond',
    'Heating', 'CentralAir', 'Electrical', 'LowQualFinSF', 'BsmtHalfBath',
    'KitchenAbvGr', 'Functional', 'GarageQual', 'GarageCond', 'PavedDrive',
    '3SsnPorch', 'ScreenPorch', 'PoolArea', 'MiscVal',
    'Exterior2nd', 'GarageYrBlt'
]

ORDINALS = ["BsmtQual", "ExterQual", "ExterCond", "FireplaceQu", "KitchenQual"]
ORDINALS_ORDERING = ['Missing', 'Po', 'Fa', 'TA', 'Gd', 'Ex']

CATEGORICAL_COLUMNS = [
   'GarageQual', 'GarageCond', 'PoolQC',
  'BsmtCond', 'Alley', 'Fence','RoofStyle', 'MiscFeature',
  'Heating', 'CentralAir', 'Electrical',
  # 'BsmtHalfBath',
  # 'KitchenAbvGr',
  'Functional', 'PavedDrive',
  'Street', 'Utilities', 'LandSlope', 'Condition1', 'Condition2', 'RoofMatl',
  'MSZoning', 'LotShape', 'LandContour', 'LotConfig', 'Neighborhood',
  'Condition1', 'BldgType', 'HouseStyle', 'RoofStyle', 'Exterior1st',
  'Exterior2nd',
  'Foundation', 'BsmtExposure', 'BsmtFinType1', 'BsmtFinType2', 'HeatingQC',
  'GarageType', 'GarageFinish', 'SaleType', 'SaleCondition',
  # "BsmtQual", "ExterQual", "ExterCond", "FireplaceQu", "KitchenQual",
  "MasVnrType"
]

NUMERICAL_COLUMNS = [
  'BsmtFinSF2', 'BsmtUnfSF', 'TotalBsmtSF', '1stFlrSF', '2ndFlrSF','GrLivArea',
  'BsmtFullBath', 'FullBath', 'HalfBath', 'BedroomAbvGr', 'TotRmsAbvGrd',
  'MiscVal',
  'GarageArea', 'WoodDeckSF', 'OpenPorchSF', 'EnclosedPorch', 'GarageCars',
  'MoSold', 'LotFrontage', 'LotArea', 'MSSubClass', 'WoodDeckSF', 'OpenPorchSF',
  'EnclosedPorch', 'MoSold', 'MasVnrArea', 'BsmtFinSF1', 'LowQualFinSF', '3SsnPorch', 'ScreenPorch'
]

# default params
DEFAULT_REDUNDANT_COLUMNS_THRESHOLD = 0.9
DEFAULT_NAN_COLUMNS_THRESHOLD = 0.9
DEFAULT_RARE_CATEGORIES_DROP_THRESHOLD = 0.1

DEFAULT_INPUT_PATH = "data/raw/train.csv"
DEFAULT_OUTPUT_PATH = "data/interim/train.csv"

# arguments helps
INPUT_PATH_HELP = "The path of the csv file to preprocess"
OUTPUT_PATH_HELP = "Where to store the processed path"
REDUNDANT_COLUMNS_HELP = """A threshold to use above which to drop drop nan columns"""
NAN_COLUMNS_THRESHOLD_HELP = """Drop if the column is mostly the same"""
RARE_CATEGORIES_DROP_THRESHOLD_HELP = "Replace rare categories"
VERBOSE_HELP = "Whether to print the steps of the pipeline"

DROP_COLUMNS = ['GarageYrBlt', 'YrSold', 'Exterior2nd', 'PoolQC']

TARGET_VARIABLE_NAME = "SalePrice"
